import unittest
from ppci.codegen.registerallocator import RegisterAllocator
from ppci.api import get_arch
from ppci.arch.arch import Frame
from ppci.arch.example import Def, Use, Add, Mov, R0, R1, ExampleRegister
from ppci.binutils.debuginfo import DebugDb


class RegisterAllocatorTestCase(unittest.TestCase):
    """ Use the example target to test the different cases of the register
        allocator.
        Possible cases: freeze of move, spill of register
    """
    def setUp(self):
        arch = get_arch('example')
        debug_db = DebugDb()
        self.register_allocator = RegisterAllocator(arch, debug_db)

    def conflict(self, ta, tb):
        color_a = self.register_allocator.Node(ta).color
        color_b = self.register_allocator.Node(tb).color
        self.assertNotEqual(color_a, color_b)

    def test_register_allocation(self):
        f = Frame('tst', [], [], None, [])
        t1 = ExampleRegister('t1')
        t2 = ExampleRegister('t2')
        t3 = ExampleRegister('t3')
        t4 = ExampleRegister('t4')
        t5 = ExampleRegister('t5')
        f.regs = [ExampleRegister('t{}'.format(v), v) for v in range(7)]
        f.instructions.append(Def(t1))
        f.instructions.append(Def(t2))
        f.instructions.append(Def(t3))
        f.instructions.append(Add(t4, t1, t2))
        f.instructions.append(Add(t5, t4, t3))
        f.instructions.append(Use(t5))
        self.register_allocator.alloc_frame(f)
        self.conflict(t1, t2)
        self.conflict(t2, t3)

    def test_register_coalescing(self):
        """ Register coalescing happens when a move can be eliminated """
        f = Frame('tst', [], [], None, [])
        t1 = ExampleRegister('t1')
        t2 = ExampleRegister('t2')
        t3 = ExampleRegister('t3')
        t4 = ExampleRegister('t4')
        t5 = ExampleRegister('t5')
        t6 = ExampleRegister('t6')
        f.regs = [ExampleRegister('t{}'.format(v), v) for v in range(7)]
        f.instructions.append(Def(t1))
        f.instructions.append(Def(t2))
        f.instructions.append(Def(t3))
        f.instructions.append(Add(t4, t2, t1))
        f.instructions.append(Mov(t5, t3))
        f.instructions.append(Add(t5, t4, t5))
        f.instructions.append(Mov(t6, t5))
        f.instructions.append(Use(t6))
        self.register_allocator.alloc_frame(f)
        self.conflict(t1, t2)
        self.conflict(t2, t3)
        self.conflict(t1, t3)

    def test_constrained_move(self):
        """ Test a constrained move.

            Do this by creating two pre-colored registers. Move one
            and add the other with the copy of the first.

            The move can then not be coalesced, and will be frozen.
        """
        f = Frame('tst', [], [], None, [])
        t1 = R0
        t2 = R0
        t3 = ExampleRegister('t3')
        t4 = ExampleRegister('t4')
        f.regs = [ExampleRegister('R{}'.format(v), v) for v in range(3)]
        f.instructions.append(Def(t1))
        move = Mov(t3, t1, ismove=True)
        f.instructions.append(move)
        f.instructions.append(Def(t2))
        f.instructions.append(Add(t4, t2, t3))
        f.instructions.append(Use(t4))
        self.register_allocator.alloc_frame(f)

        # Check t1 and t2 are pre-colored:
        self.assertEqual(
            {self.register_allocator.Node(R0)},
            self.register_allocator.precolored)
        self.assertEqual(set(), self.register_allocator.coalescedMoves)
        self.assertEqual({move}, self.register_allocator.constrainedMoves)
        self.conflict(t2, t3)
        self.assertEqual(set(), self.register_allocator.frozenMoves)
        self.assertIn(move, f.instructions)

    def test_freeze(self):
        """ Create a situation where no select and no coalesc is possible
        """
        f = Frame('tst', [], [], None, [])
        t4 = ExampleRegister('t4')
        t5 = ExampleRegister('t5')
        f.regs = [ExampleRegister('R{}'.format(v), v) for v in range(3)]
        f.instructions.append(Def(R0))
        move = Mov(R1, R0, ismove=True)
        f.instructions.append(move)
        f.instructions.append(Def(t4))
        f.instructions.append(Add(t5, t4, R1))
        f.instructions.append(Use(t5))
        self.register_allocator.alloc_frame(f)

        self.assertEqual(set(), self.register_allocator.coalescedMoves)
        # self.assertEqual({move}, self.register_allocator.frozenMoves)
        self.conflict(R1, R0)

    def test_spill(self):
        pass


if __name__ == '__main__':
    unittest.main()
